import pandas as pd
import sys
import logging
from typing import List, Tuple, Any
from timemachines.skatertools.utilities.conventions import wrap
from timemachines.skatertools.utilities.epochtime import infer_freq_from_epoch, is_valid_freq, epoch_to_naive_datetime, EPOCH
from timemachines.skatertools.utilities.suppression import no_stdout_stderr
from timemachines.skaters.proph.prophparams import PROPHET_MODEL, PROPHET_META
from timemachines.skatertools.utilities.wrangling import transpose

try: 
    from prophet import Prophet
    using_prophet=True
except ImportError:
    using_prophet=False

logging.disable(sys.maxsize)
logging.getLogger('fbprophet').setLevel(logging.ERROR)

if using_prophet:

    # Wraps the core prophet prediction capability into a plain vanilla prediction function
    # Unlike most skaters, this "integrated" prophet skater (iskater) isn't computed by calling the skater
    # successively - the opposite is true. We fit once on the entire history every data point, and that's that.
    # Thus this function, which optionally returns the forecast dataframe and the model, doesn't contain the
    # ancillary information that the skater carries (i.e. the prediction parade etc).


    def prophet_iskater_factory(y: [[float]], k: int, a: List = None, t: List = None, e=None, freq: str = None, n_max=1000,
                                recursive: bool = False, model_params: dict = None, return_forecast=True):
        """
        :param y:           A list of observations, each a vector.
        :param k:           Number of steps ahead to predict
        :param a:           Known in advance observations - should be k more of these than y's
        :param t:           Epoch times of observations y. If len(t)=len(y)+k the last k are interpreted as future times.
        :param freq:        'D', '5T' etc, see https://github.com/pandas-dev/pandas/blob/master/pandas/tseries/frequencies.py
        :param n_max:       Maximum number of observations to use, should you wish to prevent prophet from slowing down
        :param recursive    If True, exogenous variables y[1], y[2],... will be predicted forward in time
                                 (obviously this adds to computation time)
        :returns: x         k-vector of predictions
                  x_std     k-vector of standard deviations
                  forecast  full forecast dataframe, familiar to users of fbprophet
        """
        if a:
            assert len(a) == len(y) + k

        if isinstance(y[0], float):
            y = [wrap(yj) for yj in y]

        # Conversion of epoch times to UTC datetime
        # User must supply times, len(y) or len(y)+k, or a valid frequency str
        if t is None:
            if freq is None or not freq:
                freq = PROPHET_META['freq']  # Just assume away ...
            else:
                assert is_valid_freq(freq), 'Freq ' + str(freq) + ' is not a valid frequency'
            dt = pd.date_range(start=EPOCH, periods=len(y), freq=freq)  # UTC
        else:
            freq = infer_freq_from_epoch(t)
            dt = epoch_to_naive_datetime(t)

        if len(dt) == len(y) + k:
            ta = dt
            dt = dt[:len(y)]
        else:
            assert len(dt) == len(y), 'Time vector t should be len(y) or len(y)+k'
            ta = None

        # Truncate history so that prophet doesn't take forever to fit
        y_shorter = y[-n_max:]
        a_shorter = a[-(n_max + k):] if a is not None else []  # may be empty
        dt_shorter = dt[-n_max:]

        # Massage data into Prophet friendly dataframe with columns y, y1, ..., yk, a0,...aj
        y_cols = ['y' + str(i) if i > 0 else 'y' for i in range(len(y_shorter[-1]))]
        if a:
            a_cols = ['a' + str(i) for i in range(len(a_shorter[-1]))]
            data = [list(yi) + list(ai) for yi, ai in zip(y_shorter, a_shorter[:-k])]
            df = pd.DataFrame(columns=y_cols + a_cols, data=data)
        else:
            data = [list(yi) for yi in y_shorter]
            df = pd.DataFrame(columns=y_cols, data=data)
        df['ds'] = dt_shorter

        # Instantiate Prophet model, ensure defaults are what we think they are
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)

        # Add regressors
        for y_col in y_cols[1:]:
            m.add_regressor(name=y_col)
        if a:
            for a_col in a_cols:
                m.add_regressor(name=a_col)

        # Fit the model every invocation ... there isn't any other way
        with no_stdout_stderr():
            m.fit(df)

        # Make future dataframe, adding known-in-advance variables
        future = m.make_future_dataframe(periods=k, freq=freq)
        if a:
            for j, a_col in enumerate(a_cols):
                future[a_col] = [ai[j] for ai in a_shorter]  # Known in advance
        if ta is not None:
            future['ds'] = ta  # override with user supplied future times

        # Next, we wish to add contemporaneously observed variables
        #
        # This is somewhat problematic, for how should we bring exogenously observed variables forward?
        # The simplest answer is, don't use them - only supply 1-vector y observations
        # prophet implicitly assumes all exogenous are known, which is a pretty big shortcoming.
        #
        # However, if we are trying to support y[1:], ...
        #   - It seems consistent to use prophet to predict these forward,
        #   - It also seems likely that this will lead to over-fitting.
        # I'm open to ideas here. Perhaps perform some hackery could effect attenuation of the coefficients
        # assigned to y[1],... such as jiggling past observations. For now we use prophet on each
        # one individually, feeding them the known in advance 'a' variables.

        n_exog = len(y[0]) - 1
        if n_exog > 0:
            for j,y_col in enumerate(y_cols):
                if j>0:
                    yj = [yi[j] for yi in y_shorter]
                    if recursive:
                        yj_hat, yj_hat_std, yj_forecast, yj_m = prophet_iskater_factory(y=yj, k=k, a=a_shorter, freq=freq,
                                                                                        n_max=n_max, recursive=False)
                    else:
                        yj_hat = [yj[-1]] * k
                    future[y_col] = yj + list(yj_hat)

        # Call the prediction function
        forecast = m.predict(future)
        x = list(forecast['yhat'].values[-k:])  # Use m.plot(forecast) to take a peak

        # Interpret confidence level difference as scale to be returned. TODO: set alpha properly so this really is 1-std
        x_std = list([u - l for u, l in zip(forecast['yhat_upper'].values[-k:], forecast['yhat_lower'].values[-k:])])

        if return_forecast:
            return x, x_std, forecast, m
        else:
            return x, x_std


    # The rest of this module contains redundant special cases of the above, purely for offlinetesting.

    def prophet_fit_and_predict_simple(y: [float], k: int, freq: str = None, model_params: dict = None) -> Tuple[
        List, List, Any, Any]:
        """ Simpler wrapper for offlinetesting - univariate only """

        df = pd.DataFrame(columns=['y'], data=y)
        freq = freq or PROPHET_META['freq']
        df['ds'] = pd.date_range(start=EPOCH, periods=len(y), freq=freq)  # UTC
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)
        with no_stdout_stderr():
            m.fit(df)
        future = m.make_future_dataframe(periods=k, freq=freq)
        forecast = m.predict(future)
        x = forecast['yhat'].values[-k:]  # Use m.plot(forecast) to take a peak
        x_std = [u - l for u, l in zip(forecast['yhat_upper'].values, forecast['yhat_lower'].values)]
        return x, x_std, forecast, m


    def prophet_fit_and_predict_with_time(y: [float], k: int, t: [float], model_params: dict = None) -> Tuple[
        List, List, Any, Any]:
        """ Simpler wrapper for offlinetesting - univariate only w/ supplied times """

        df = pd.DataFrame(columns=['y'], data=y)
        df['ds'] = epoch_to_naive_datetime(t)
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)
        with no_stdout_stderr():
            m.fit(df)
        freq = infer_freq_from_epoch(t)
        future = m.make_future_dataframe(periods=k, freq=freq)
        forecast = m.predict(future)
        x = forecast['yhat'].values[-k:]  # Use m.plot(forecast) to take a peak
        x_std = [u - l for u, l in zip(forecast['yhat_upper'].values, forecast['yhat_lower'].values)]
        return x, x_std, forecast, m


    def prophet_fit_and_predict_with_time_and_advance_time(y: [float], k: int, t: [float], model_params: dict = None) -> \
    Tuple[List, List, Any, Any]:
        """ Simpler wrapper for offlinetesting - univariate only w/ supplied times and future times  """
        assert len(t) == len(y) + k
        df = pd.DataFrame(columns=['y'], data=y)
        dt = epoch_to_naive_datetime(t)
        df['ds'] = dt[:len(y)]
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)
        with no_stdout_stderr():
            m.fit(df)
        freq = infer_freq_from_epoch(t)
        future = m.make_future_dataframe(periods=k, freq=freq)
        future['ds'] = dt
        forecast = m.predict(future)
        x = forecast['yhat'].values[-k:]  # Use m.plot(forecast) to take a peak
        x_std = [u - l for u, l in zip(forecast['yhat_upper'].values, forecast['yhat_lower'].values)]
        return x, x_std, forecast, m


    def prophet_fit_and_predict_with_advance_vars(y: [float], k: int, t: [float], a: [[float]],
                                                  model_params: dict = None) -> Tuple[List, List, Any, Any]:
        """ Simpler wrapper for offlinetesting - univariate w/ advance vars w/ supplied times and future times  """
        assert len(t) == len(y) + k
        assert len(a) == len(y) + k
        assert isinstance(y[0], float)
        a_cols = ['a' + str(i) for i in range(len(a[0]))]
        df = pd.DataFrame(columns=a_cols, data=a[:-k])
        df['y'] = y
        dt = epoch_to_naive_datetime(t)
        df['ds'] = dt[:len(y)]
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)
        for a_col in a_cols:
            m.add_regressor(name=a_col)
        with no_stdout_stderr():
            m.fit(df)
        freq = infer_freq_from_epoch(t)
        future = m.make_future_dataframe(periods=k, freq=freq)
        future['ds'] = dt
        full_a_data = transpose(a)
        for a_col, a_vals in zip(a_cols, full_a_data):
            future[a_col] = a_vals
        forecast = m.predict(future)
        x = forecast['yhat'].values[-k:]  # Use m.plot(forecast) to take a peak
        x_std = [u - l for u, l in zip(forecast['yhat_upper'].values, forecast['yhat_lower'].values)]
        return x, x_std, forecast, m


    def prophet_fit_and_predict_with_exog_and_advance_vars(y: [[float]], k: int, t: [float], a: [[float]],
                                                           model_params: dict = None) -> Tuple[List, List, Any, Any]:
        """ Simpler wrapper for offlinetesting - univariate w/ advance vars w/ supplied times and future times  """
        assert len(t) == len(y) + k
        assert len(a) == len(y) + k
        assert isinstance(y[0], List)
        a_cols = ['a' + str(i) for i in range(len(a[0]))]
        df = pd.DataFrame(columns=a_cols, data=a[:-k])
        Y = transpose(y)
        df['y'] = Y[0]
        n_exog = len(y[0]) - 1
        y_cols = ['y'] + ['y' + str(i) for i in range(1, len(y[0]))]
        for i in range(1, n_exog + 1):
            df['y' + str(i)] = Y[i][:len(y)]
        dt = epoch_to_naive_datetime(t)
        df['ds'] = dt[:len(y)]
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)
        for a_col in a_cols:
            m.add_regressor(name=a_col)
        for y_col in y_cols[1:]:
            m.add_regressor(name=y_col)
        with no_stdout_stderr():
            m.fit(df)
        freq = infer_freq_from_epoch(t)
        future = m.make_future_dataframe(periods=k, freq=freq)
        future['ds'] = dt
        full_a_data = transpose(a)
        for a_col, a_vals in zip(a_cols, full_a_data):
            future[a_col] = a_vals
        for i in range(1, n_exog + 1):  # Just bring forward
            future['y' + str(i)] = Y[i] + [Y[i][-1]] * k
        forecast = m.predict(future)
        x = forecast['yhat'].values[-k:]  # Use m.plot(forecast) to take a peak
        x_std = [u - l for u, l in zip(forecast['yhat_upper'].values, forecast['yhat_lower'].values)]
        return x, x_std, forecast, m


    def prophet_fit_and_predict_with_exog_and_advance_vars_no_t(y: [[float]], k: int, freq: str, a: [[float]],
                                                                model_params: dict = None) -> Tuple[List, List, Any, Any]:
        """ Simpler wrapper for offlinetesting - univariate w/ advance vars w/ supplied times and future times  """
        assert len(a) == len(y) + k
        assert isinstance(y[0], List)
        a_cols = ['a' + str(i) for i in range(len(a[0]))]
        df = pd.DataFrame(columns=a_cols, data=a[:-k])
        Y = transpose(y)
        df['y'] = Y[0]
        n_exog = len(y[0]) - 1
        y_cols = ['y'] + ['y' + str(i) for i in range(1, len(y[0]))]
        for i in range(1, n_exog + 1):
            df['y' + str(i)] = Y[i][:len(y)]
        df['ds'] = pd.date_range(EPOCH, periods=len(y), freq=freq)
        kwargs_used = dict([(k, v) for k, v in PROPHET_MODEL.items()])
        if model_params:
            kwargs_used.update(model_params)
        m = Prophet(**kwargs_used)
        for a_col in a_cols:
            m.add_regressor(name=a_col)
        for y_col in y_cols[1:]:
            m.add_regressor(name=y_col)
        with no_stdout_stderr():
            m.fit(df)
        future = m.make_future_dataframe(periods=k, freq=freq)
        full_a_data = transpose(a)
        for a_col, a_vals in zip(a_cols, full_a_data):
            future[a_col] = a_vals
        for i in range(1, n_exog + 1):  # Just bring forward
            future['y' + str(i)] = Y[i] + [Y[i][-1]] * k
        forecast = m.predict(future)
        x = forecast['yhat'].values[-k:]  # Use m.plot(forecast) to take a peak
        x_std = [u - l for u, l in zip(forecast['yhat_upper'].values, forecast['yhat_lower'].values)]
        return x, x_std, forecast, m




import numpy as np
from timemachines.skatertools.data.synthetic import brownian_with_exogenous, brownian_with_noise
from timemachines.skating import prior
from timemachines.skatertools.utilities.conventions import wrap

# Quick ways to inspect skaters

try:
    import matplotlib.pyplot as plt
    using_matplotlib = True
except ImportError:
    using_matplotlib = False
    plt = None


def prior_plot(f, y=None, k=1, t=None, e=None, r=None, x0=np.nan, n=150, n_plot=25):
    """
         Apply state machine to univariate series,
         Show observations and out of sample predictions predictions
    """
    assert using_matplotlib, 'pip install matplotlib'

    if y is None:
       y = brownian_with_noise(n=n)

    if t is None:
        t = [float(ti) for ti in range(len(y))]

    x, x_std = prior(f=f, y=y, k=k, a=t, t=t, e=e, r=r, x0=x0)
    ysf = [[wrap(y_)[0]] for y_ in y]
    xk = [xt[-1] for xt in x]
    plot_with_last_value(t=t, x=xk, y=ysf, k=k, n_plot=n_plot)


def prior_plot_exogenous(f, y=None, k=None, a=None, t=None, e=None, r=None, x0=np.nan, n=150, n_plot=25):
    """
          Apply state machine to univariate series,
          Show observations, out of sample predictions predictions, and exogenous variables
    """
    assert using_matplotlib, 'pip install matplotlib'

    if y is None:
       y = brownian_with_exogenous(n)

    if t is None:
       t = range(len(y))

    e = [ None]*(len(y)-n_plot) + [10]*n_plot

    x,x_std = prior(f=f, y=y, k=k, a=a, t=t, e=e, r=r, x0=x0)
    xk = [xt[-1] for xt in x]
    plot_with_last_value(t=t, x=xk, y=y, k=k, n_plot=n_plot)


def plot_with_last_value(t, x, y, k, n_plot:int):
    """
    :param t:        Time of observation
    :param x:        Point estimates
    :param y:        Univariate or Multivariate series of observations
    :param n_plot:    Number of examples to plot (from end)
    :return:
    """
    assert using_matplotlib, 'pip install matplotlib'
    assert isinstance(x[0],float)
    try:
        y0 = [y_[0] for y_ in y]
    except:
        y0 = y
    plt.plot(t[-n_plot:], y0[-n_plot:], 'b*')
    lgnd = ['target']
    for j in range(1, len(y[0])):
        ysj = [y_[j] for y_ in y]
        plt.plot(t[-n_plot:], ysj[-n_plot:], 'r+')
        lgnd.append('exogenous '+str(j))

    lv_x = [y0[0]]*k + y0
    lv_x = lv_x[:len(t)]

    plt.plot(t[-n_plot:], x[-n_plot:], 'g-')
    lgnd.append('prediction')

    plt.plot(t[-n_plot:], lv_x[-n_plot:],'g--')
    lgnd.append('last_value')

    plt.legend(lgnd)
    plt.show()





from timemachines.skaters.orbt.orbitinclusion import using_orbit

if using_orbit:
    from timemachines.skaters.orbt.orbitlgtskaterfactory import orbit_lgt_12, orbit_lgt_24
    ORBIT_LGT_SKATERS = [orbit_lgt_12, orbit_lgt_24]
else:
    ORBIT_LGT_SKATERS = []


ORBIT_SKATERS = ORBIT_LGT_SKATERS


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="timemachines",
    version="0.13.14",
    description="Evaluation and standardization of popular time series packages",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/microprediction/timemachines",
    author="microprediction",
    author_email="pcotton@intechinvestments.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["timemachines",
              "timemachines.skaters",
              "timemachines.skaters.divine",
              "timemachines.skaters.dlm",
              "timemachines.skaters.flux",
              "timemachines.skaters.pmd",
              "timemachines.skaters.tsa",
              "timemachines.skaters.proph",
              "timemachines.skaters.nproph",
              "timemachines.skaters.simple",
              "timemachines.skaters.orbt",
              "timemachines.skaters.bats",
              "timemachines.skaters.rvr",
              "timemachines.skaters.sk",
              "timemachines.skaters.elo",
              "timemachines.skaters.smdk",
              "timemachines.skaters.gk",
              "timemachines.skaters.drts",
              "timemachines.skatertools",
              "timemachines.skatertools.comparison",
              "timemachines.skatertools.components",
              "timemachines.skatertools.composition",
              "timemachines.skatertools.ensembling",
              "timemachines.skatertools.data",
              "timemachines.skatertools.evaluation",
              "timemachines.skatertools.tuning",
              "timemachines.skatertools.utilities",
              "timemachines.skatertools.visualization",
              "timemachines.skatertools.batch",
              "timemachines.skatertools.recommendations"],
    test_suite='pytest',
    tests_require=['pytest','microprediction'],
    include_package_data=True,
    install_requires=["wheel","pathlib","numpy>=1.19.5","pandas","importlib-metadata>=1.7.0",
                      "microconventions>0.5.0","getjson","lunarcalendar","holidays",
                      "sklearn","scipy","statsmodels>=0.12.2","river","convertdate>=2.2.0",
                      "momentum>=0.1.2","requests"],
    entry_points={
        "console_scripts": [
            "timemachines=timemachines.__main__:main",
        ]
    },
)



from timemachines.skaters.elo.alleloskaters import ELO_SKATERS

NETWORKED_SKATERS = ELO_SKATERS
FAST_NETWORKED_SKATERS = ELO_SKATERS

# Networked skaters access Elo ratings


def networked_skater_from_name(name):
    valid = [f for f in NETWORKED_SKATERS if f.__name__ == name]
    return valid[0] if len(valid)==1 else None



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
from timemachines.skatertools.components.parade import parade
from timemachines.skatertools.utilities.nonemath import nonecast, nonecenter
from timemachines.skaters.bats.batsinclusion import using_bats
from timemachines.skatertools.utilities.conventions import wrap
from timemachines.skaters.bats.batsifactory import bats_iskater_factory
import numpy as np
from timemachines.skatertools.batch.batchskater import batch_skater_factory


if using_bats:
    from tbats import TBATS

    def bats_factory(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None,
                     emp_mass: float = 0.0,
                     use_box_cox=None, box_cox_bounds=(0, 1),
                     use_trend=None, use_damped_trend=None,
                     seasonal_periods=None, use_arma_errors=True
                     ):
        return batch_skater_factory(y=y,s=s,k=k,a=a,t=t,e=e,r=r,
                                    iskater=bats_iskater_factory,
                                    iskater_kwargs={'use_box_cox':use_box_cox,
                                                    'box_cox_bounds':box_cox_bounds,
                                                    'use_trend':use_trend,
                                                    'use_damped_trend':use_damped_trend,
                                                     'seasonal_periods':seasonal_periods,
                                                    'use_arma_errors':use_arma_errors},
                                    min_e=0, emp_mass=emp_mass, emp_std_mass=1.0, n_warm=10)


    def bats_factory_old(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None,
                     emp_mass: float = 0.0,
                     use_box_cox=None, box_cox_bounds=(0, 1),
                     use_trend=None, use_damped_trend=None,
                     seasonal_periods=None, use_arma_errors=True
                     ):
        """
          :param emp_mass:        How much to use empirical bias to correct
        """
        assert 0<= emp_mass<=1

        if e is None:
           e = 1  # 'Providing e=None to bats is not recommended. Use e<0 to skip a fitting or e>0 to fit.'

        y = wrap(y)
        a = wrap(a)

        if not s.get('y'):
            s = {'p': {},      # parade
                 'y': list(),  # historical y
                 'a': list(),  # list of a known k steps in advance
                 't': list(),
                 'k': k}
        else:
            # Assert immutability of k, dimensions of y,a
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']
            if s['a']:
                assert len(a) == len(s['a'][0])

        if y is None:
            return None, s, None
        else:
            s['y'].append(y)
            if a is not None:
                s['a'].append(a)
            if t is not None:
                assert isinstance(t,float), 'epoch time please'
                s['t'].append(t)

            if (e is not None) and (e>0) and ((a is None) or (len(s['a'])>k)) and (len(s['y'])>10):
                # Offset y, t, a are supplied to prophet interface
                # Prophet requires several non-nan rows
                t_arg = s['t'][k:] if t is not None else None
                a_arg = s['a']
                y_arg = s['y'][k:]
                x, x_std = bats_iskater_factory(y=y_arg, k=k, a=a_arg, t=t_arg,
                                                                 use_box_cox=use_box_cox, box_cox_bounds=box_cox_bounds,
                                                                 use_trend=use_trend, use_damped_trend=use_damped_trend,
                                                                 seasonal_periods=seasonal_periods, use_arma_errors=use_arma_errors
                                                                 )
                s['m'] = True # Flag indicating a model has been fit (there is no point keeping the model itself, however)
            else:
                x = [y[0]] * k
                x_std = [1] * k

            # Get running mean prediction errors from the prediction parade
            x_resid, x_resid_std, s['p'] = parade(p=s['p'], x=x, y=y[0])
            x_resid_std = nonecast(x_resid_std,1.0)

            # Compute center of mass between bias-corrected and uncorrected predictions
            emp_std_mass = 1.0 # <--- Makes no sense to use x_std_rubbish as the model doesn't compute it
            x_corrected = np.array(x_resid) + np.array(x)
            x_center = nonecenter(m=[emp_mass, 1 - emp_mass], x=[x_corrected, x])
            x_std_center = nonecenter(m=[emp_std_mass, 1 - emp_std_mass], x=[x_resid_std, x_std])

            return x_center, x_std_center, s


from timemachines.skaters.nproph.nprophetskaters import NPROPHET_UNIVARIATE_SKATERS
from timemachines.skaters.nproph.nprophskaterscomposed import NPROPHET_SKATERS_COMPOSED

NPROPHET_SKATERS = NPROPHET_UNIVARIATE_SKATERS + NPROPHET_SKATERS_COMPOSED


from timemachines.skaters.pmd.pmdinclusion import using_pmd
if using_pmd:
    from timemachines.skaters.pmd.pmdskaters import pmd_known, pmd_exogenous, pmd_univariate
    from timemachines.skaters.pmd.pmdcomposed import pmd_exogenous_hypocratic, pmd_univariate_hypocratic

    PMD_SKATERS = [pmd_univariate, pmd_exogenous_hypocratic]  # for now
else:
    PMD_SKATERS = []


from timemachines.skaters.rvr.rvrinclusion import using_river

if using_river:
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
    from timemachines.skaters.rvr.rvrconstant import RIVER_CONSTANT_SKATERS
    from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
    from timemachines.skatertools.utilities.conventions import wrap


    def hypocratic_rvr_factory(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None):
        """
             r :  moving average parameter  (e.g. 0.75 is fast, 0.95 is slow)
        """
        y0 = wrap(y)[0]
        assert r is not None
        x, x_std, s = precision_weighted_ensemble_factory(fs=RIVER_CONSTANT_SKATERS, y=y0, s=s, k=k, a=a, t=t, e=e, r=r)

        def hypocratic(x:float, x_std:float, confidence=0.5):
            """ Shrink residual prediction towards zero """
            # Remains to be seen if this is sensible for TSA residuals
            import math
            if abs(x_std)<1e-6 or abs(x)<1e-3*x_std:
                return 0.0
            else:
                return x*math.tanh(confidence*abs(x)/(3*x_std))

        x_resid = [ hypocratic(xi,x_std) for xi,x_std in zip(x,x_std) ]
        return x_resid, x_std, s


    def rvr_slowly_hypocratic(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
        return hypocratic_rvr_factory(y=y,s=s,k=k,a=a,t=t,e=e,r=0.95)


    def rvr_quickly_hypocratic(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
        return hypocratic_rvr_factory(y=y,s=s,k=k,a=a,t=t,e=e,r=0.75)


    RIVER_HYPOCRATIC_SKATERS = [rvr_quickly_hypocratic, rvr_slowly_hypocratic ]
else:
    RIVER_HYPOCRATIC_SKATERS = []


from timemachines.skaters.sk.skinclusion import using_sktime
from timemachines.skaters.pmd.pmdinclusion import using_pmd

if using_sktime and using_pmd:
    from timemachines.skaters.sk.skwrappers import sk_autoarima_iskater
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.batch.batchskater import batch_skater_factory

    def sk_autoarima(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=sk_autoarima_iskater,
                                    iskater_kwargs={},
                                    min_e=0)

    SK_AA_SKATERS = [sk_autoarima]

else:
    SK_AA_SKATERS = []

if __name__=='__main__':
    from timemachines.skatertools.evaluation.evaluators import hospital_mean_square_error_with_sporadic_fit
    err1 = hospital_mean_square_error_with_sporadic_fit(f=sk_autoarima, k=3, n=110)



from timemachines.skaters.smdk.smdkinclusion import using_simdkalman

if using_simdkalman:
    from timemachines.skaters.smdk.smdkarmafactory import smdk_arma_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap

    def smdk_p5_d0_q3_n1000(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
          return smdk_arma_factory(y=y, n_agents=1000, max_p=5, max_q=3, s=s, k=k, a=a, t=t, e=e, r=0.75 )


    def smdk_p5_d0_q3_n1000_aggressive(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                            r: R_TYPE = None):
        return smdk_arma_factory(y=y, n_agents=1000, max_p=5, max_q=3, s=s, k=k, a=a, t=t, e=e, r=0.95)


    def smdk_p5_d0_q3_n500(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return smdk_arma_factory(y=y, n_agents=500, max_p=5, max_q=3, s=s, k=k, a=a, t=t, e=e, r=0.75)


    def smdk_p5_d0_q3_n500_aggressive(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return smdk_arma_factory(y=y, n_agents=500, max_p=5, max_q=3, s=s, k=k, a=a, t=t, e=e, r=0.95)


    def smdk_p5_d0_q3_n5000(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                           r: R_TYPE = None):
        return smdk_arma_factory(y=y, n_agents=5000, max_p=5, max_q=3, s=s, k=k, a=a, t=t, e=e, r=0.75)


    SMDK_ARMA_SKATERS = [smdk_p5_d0_q3_n1000, smdk_p5_d0_q3_n500, smdk_p5_d0_q3_n500_aggressive, smdk_p5_d0_q3_n1000_aggressive]
else:
    SMDK_ARMA_SKATERS = []



if __name__=='__main__':
    from timemachines.skatertools.data.real import hospital_with_exog
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error_with_sporadic_fit
    from timemachines.skaters.simple.thinking import thinking_fast_and_slow
    import time
    k = 1
    st = time.time()
    y, a = hospital_with_exog(k=k, n=500, offset=True)
    f = smdk_p5_d0_q3_n1000
    err = evaluate_mean_squared_error_with_sporadic_fit(f,y=y,k=k, fit_frequency=1)
    elapsed = time.time() - st
    st = time.time()
    err0 = evaluate_mean_squared_error_with_sporadic_fit(f=thinking_fast_and_slow,y=y,k=k,fit_frequency=1)
    elapsed0 = time.time()-st
    n_calcs = (len(y)-10-k)*1000
    print('Relative CPU = '+str(elapsed/elapsed0)+' with '+str(round(n_calcs/elapsed))+' agent forecasts per second for error of '+str(err/err0))
    if False:
        from timemachines.skatertools.visualization.priorplot import prior_plot
        prior_plot(f=f, k=k, y=y, n=450, n_plot=25)

